# Copyright (c) 2021 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from sudachipy.dictionary import Dictionary
from sudachipy.morphemelist import MorphemeList
from sudachipy.tokenizer import Tokenizer


class SudachipyWordTokenizer:
    """Runs tokenization with SudachiPy."""

    def __init__(
            self,
            split_mode: Optional[str] = "C",
            config_path: Optional[str] = None,
            resource_dir: Optional[str] = None,
            dict_type: Optional[str] = "core",
    ):
        """
        Constructs a SudachipyTokenizer.

        Args:
            split_mode (:obj:`str`, `optional`, defaults to :obj:`"C"`):
                The mode of splitting.
                "A", "B", or "C" can be specified.
            config_path (:obj:`str`, `optional`, defaults to :obj:`None`):
                Path to a config file of SudachiPy to be used for the sudachi dictionary initialization.
            resource_dir (:obj:`str`, `optional`, defaults to :obj:`None`):
                Path to a resource dir containing resource files, such as "sudachi.json".
            dict_type (:obj:`str`, `optional`, defaults to :obj:`"core"`):
                Sudachi dictionary type to be used for tokenization.
                "small", "core", or "full" can be specified.
        """
        split_mode = split_mode.upper()
        if split_mode == "C":
            self.split_mode = Tokenizer.SplitMode.C
        elif split_mode == "B":
            self.split_mode = Tokenizer.SplitMode.B
        elif split_mode == "A":
            self.split_mode = Tokenizer.SplitMode.A
        else:
            raise ValueError("Invalid `split_mode`: " + split_mode)

        sudachi_dict = Dictionary(config_path=config_path, resource_dir=resource_dir, dict_type=dict_type)
        self.sudachi = sudachi_dict.create()

    def tokenize(self, text: str) -> MorphemeList:
        """
        Tokenizes the specified text and returns its morphemes.

        Args:
            text (str): Input string.

        Returns:
            MorphemeList: List of morphemes.
        """
        return self.sudachi.tokenize(text, self.split_mode)



# Copyright (c) 2021 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import textspan
from tokenizers import NormalizedString, PreTokenizedString
from typing import List, Optional

from .. import SudachipyWordTokenizer
from ..tokenization_bert_sudachipy import WORD_FORM_TYPES


class CustomPreTokenizer:
    def custom_split(self, i: int, normalized_string: NormalizedString) -> List[NormalizedString]:
        """
        Tokenizes the input string and returns list of tokens.

        Please override this function with your custom tokenizer.
        Example. https://github.com/huggingface/tokenizers/blob/b24a2fc/bindings/python/examples/custom_components.py

        Args:
            i (int): Index.
            normalized_string (NormalizedString): Input string.

        Returns:
            List[NormalizedString]: List of normalized_strings.
        """
        raise NotImplementedError()

    def pre_tokenize(self, pretok: PreTokenizedString):
        pretok.split(self.custom_split)

    @staticmethod
    def split_normalized_string(normalized_string: NormalizedString, tokens: List[str]) -> List[NormalizedString]:
        """
        Splits normalized_string by tokens.

        Args:
            normalized_string (NormalizedString): Input string.
            tokens (List[str]): List of surface words in a sentence.

        Returns:
            List[NormalizedString]: List of normalized_strings.
        """
        token_spans = textspan.get_original_spans(tokens, str(normalized_string).strip())
        return [normalized_string[start:end] for token_span in token_spans for start, end in token_span]


class SudachipyPreTokenizer(SudachipyWordTokenizer, CustomPreTokenizer):
    def __init__(
        self,
        split_mode: Optional[str] = "C",
        dict_type: Optional[str] = "core",
        word_form_type: Optional[str] = "surface",
        **kwargs
    ):
        """
        Constructs a SudachipyPreTokenizer.

        Args:
            split_mode (:obj:`str`, `optional`, defaults to :obj:`"C"`):
                The mode of splitting.
                "A", "B", or "C" can be specified.
            dict_type (:obj:`str`, `optional`, defaults to :obj:`"core"`):
                Sudachi dictionary type to be used for tokenization.
                "small", "core", or "full" can be specified.
            word_form_type (:obj:`str`, `optional`, defaults to :obj:`"surface"`):
                Word form type for each morpheme.
                "surface", "dictionary", "normalized", "dictionary_and_surface", or "normalized_and_surface" can be specified.
            **kwargs:
                Sudachi dictionary parameters.
        """
        SudachipyWordTokenizer.__init__(self, split_mode=split_mode, dict_type=dict_type, **kwargs)
        self.word_form_type = word_form_type
        self.word_formatter = WORD_FORM_TYPES[self.word_form_type] if self.word_form_type != "surface" else None

    def custom_split(self, i: int, normalized_string: NormalizedString) -> List[NormalizedString]:
        """
        Tokenizes with SudachiPy and returns its tokens.

        Args:
            i (int): Index.
            normalized_string (NormalizedString): Input string.

        Returns:
            List[NormalizedString]: List of normalized_strings.
        """
        morphs = super().tokenize(str(normalized_string).strip())
        tokens = list(map(lambda m: m.surface(), morphs))
        normalized_strings = self.split_normalized_string(normalized_string, tokens)
        if not (len(morphs) == len(tokens) == len(normalized_strings)):
            raise ValueError(len(morphs), len(tokens), len(normalized_strings), tokens, normalized_strings)

        if self.word_form_type != 'surface':
            _ = [ns.replace(ns.normalized, self.word_formatter(m)) for ns, m in zip(normalized_strings, morphs)]

        return normalized_strings



# Copyright (c) 2021 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from typing import Dict, List, Optional, Tuple, Union

from logzero import logger
from progressbar import progressbar as tqdm

from .. import SudachipyWordTokenizer
from ..tokenization_bert_sudachipy import pos_substitution_format, WORD_FORM_TYPES


class PartOfSpeechSubstitutionTokenizer(SudachipyWordTokenizer):
    def __init__(
            self,
            split_mode: Optional[str] = "C",
            dict_type: Optional[str] = "core",
            word_form_type: Optional[str] = "surface",
            **kwargs
    ):
        """
        Constructs a PartOfSpeechSubstitutionTokenizer.

        Args:
            split_mode (:obj:`str`, `optional`, defaults to :obj:`"C"`):
                The mode of splitting.
                "A", "B", or "C" can be specified.
            dict_type (:obj:`str`, `optional`, defaults to :obj:`"core"`):
                Sudachi dictionary type to be used for tokenization.
                "small", "core", or "full" can be specified.
            word_form_type (:obj:`str`, `optional`, defaults to :obj:`"surface"`):
                Word form type for each morpheme.
                "surface", "dictionary", "normalized", "dictionary_and_surface", or "normalized_and_surface" can be specified.
            **kwargs:
                Sudachi dictionary parameters.
        """
        super().__init__(split_mode=split_mode, dict_type=dict_type, **kwargs)
        self.word_form_type = word_form_type
        self.word_formatter = WORD_FORM_TYPES[self.word_form_type]
        self._vocab = None

    def get_word2freq_and_pos(self, files: List[str]) -> Tuple[Dict[str, int], List[str]]:
        """
        Tokenizes sentences in the specified files and returns tokenized data.

        Args:
            files (List[str]): List of paths of input files.

        Returns:
            Tuple[Dict[str, int], List[str]]:
                1. Dictionary of tokens and its frequency.
                2. List of part-of-speech tags.

        """
        word2freq = defaultdict(int)
        pos_list = []

        logger.info("Tokenization")
        for file in files:
            logger.info("\tReading file: {}".format(file))
            with open(file, 'r') as f:
                for line in tqdm(f):
                    line = line.strip()
                    if line != "":
                        for m in self.tokenize(line):
                            word2freq[self.word_formatter(m)] += 1
                            pos_list.append(pos_substitution_format(m))

        return word2freq, list(set(pos_list))

    def train(
            self,
            files: Union[str, List[str]],
            token_size: int = 32000,
            min_frequency: int = 1,
            limit_character: int = 1000,
            special_tokens: List[str] = ["[PAD]", "[UNK]", "[CLS]", "[SEP]", "[MASK]"]
    ):
        """
        Reads the input files and builds vocabulary.

        Args:
            files (str | List[str]):
                List of paths of input files to build vocabulary.
            token_size (int):
                The size of the vocabulary, excluding special tokens and pos tags.
            min_frequency (int):
                Ignores all words (tokens and characters) with total frequency lower than this.
            limit_character (int):
                The maximum different characters to keep in the vocabulary.
            special_tokens (List[str]):
                A list of special tokens the model should know of.
        """
        logger.info("Parameters for training")
        logger.info("\ttoken_size: {}".format(token_size))
        logger.info("\tmin_frequency: {}".format(min_frequency))
        logger.info("\tlimit_character: {}".format(limit_character))
        logger.info("\tspecial_tokens: {}".format(",".join(special_tokens)))

        if isinstance(files, str):
            files = [files]
        word2freq, pos_list = self.get_word2freq_and_pos(files)

        word2freq = {word: freq for word, freq in word2freq.items() if min_frequency <= freq}

        char2freq = {word: freq for word, freq in word2freq.items() if len(word) == 1}
        if limit_character < len(char2freq):
            sorted_char2freq = sorted(char2freq.items(), key=lambda x: x[1], reverse=True)
            for char, _ in sorted_char2freq[limit_character:]:
                del word2freq[char]

        word2freq = sorted(word2freq.items(), key=lambda x: x[1], reverse=True)
        if token_size < len(word2freq):
            word2freq = word2freq[:token_size]

        self._vocab = special_tokens + pos_list + list(map(lambda x: x[0], word2freq))
        logger.info("#Vocab, including POS, all (special) tokens and characters\n{}".format(len(self.vocab)))

    def save_vocab(self, output_path: str):
        """
        Saves Vocabulary into the specified path.

        Args:
            output_path (str): The output path where the vocabulary will be stored.
        """
        logger.info("Saving vocab to `{}`".format(output_path))
        with open(output_path, 'w') as f:
            f.write("\n".join(self.vocab))

    @property
    def vocab(self):
        return self._vocab



# Copyright (c) 2021 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
from glob import glob

from sudachitra.pretokenizer import PartOfSpeechSubstitutionTokenizer


def main():
    args = get_args()

    if args.input_file:
        files = [args.input_file]
    elif args.input_dir:
        files = glob(os.path.join(args.input_dir, '*.txt'))
    else:
        raise ValueError("`input_file` or `input_dir` must be specified.")

    pos_tokenizer = PartOfSpeechSubstitutionTokenizer(
        split_mode=args.split_mode,
        dict_type=args.dict_type,
        word_form_type=args.word_form_type
    )

    pos_tokenizer.train(
        files,
        token_size=args.token_size,
        min_frequency=args.min_frequency,
        limit_character=args.limit_character,
        special_tokens=args.special_tokens
    )

    pos_tokenizer.save_vocab(args.output_file)


def get_args():
    parser = argparse.ArgumentParser(description='Trainer of part-of-speech substitution tokenizer.')

    # Input
    parser.add_argument('-f', '--input_file', default='',
                        help='Input file to train tokenizer.')
    parser.add_argument('-d', '--input_dir', default='',
                        help='Input directory containing files to train tokenizer.')

    # Parameters
    parser.add_argument('--token_size', type=int, default=32000,
                        help='The size of the vocabulary, excluding special tokens and pos tags.')
    parser.add_argument('--min_frequency', type=int, default=1,
                        help='Ignores all words (tokens and characters) with total frequency lower than this.')
    parser.add_argument('--limit_character', type=int, default=1000,
                        help='The maximum different characters to keep in the vocabulary.')
    parser.add_argument('--special_tokens', nargs='*', default=["[PAD]", "[UNK]", "[CLS]", "[SEP]", "[MASK]"],
                        help='A list of special tokens the model should know of.')

    # Tokenization
    parser.add_argument('--dict_type', default='core', choices=['small', 'core', 'full'],
                        help='Sudachi dictionary type to be used for tokenization.')
    parser.add_argument('--split_mode', default='C', choices=['A', 'B', 'C', 'a', 'b', 'c'],
                        help='The mode of splitting.')
    parser.add_argument('--word_form_type', default='surface',
                        choices=['surface', 'dictionary', 'normalized', 'dictionary_and_surface', 'normalized_and_surface'],
                        help='Word form type for each morpheme.')

    # output
    parser.add_argument('-o', '--output_file',
                        help='The output path where the vocabulary will be stored.')

    args = parser.parse_args()

    return args


if __name__ == '__main__':
    main()



# coding=utf-8
# Copyright 2018 The HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This script can be used to convert a head-less TF2.x Bert model to PyTorch,
as published on the official GitHub: https://github.com/tensorflow/models/tree/master/official/nlp/bert

TF2.x uses different variable names from the original BERT (TF 1.4) implementation.
The script re-maps the TF2.x Bert weight names to the original names, so the model can be imported with Huggingface/transformer.

You may adapt this script to include classification/MLM/NSP/etc. heads.
"""
import argparse
import logging
import os
import re

import tensorflow as tf
import torch

from transformers import BertConfig, BertForPreTraining


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def load_tf2_weights_in_bert(model, tf_checkpoint_path, config):
    tf_checkpoint_path = os.path.abspath(tf_checkpoint_path)

    logger.info("Converting TensorFlow checkpoint from {}".format(tf_checkpoint_path))

    for full_name, shape in tf.train.list_variables(tf_checkpoint_path):
        pointer = model
        trace = []

        if "optimizer/" in full_name:
            logger.info("Skipping optimizer weights: {}".format(full_name))
            continue

        split_name = full_name.split("/")
        name = split_name.pop(0)
        if name != "model":
            logger.info("Skipping: {}".format(full_name))
            continue

        name = split_name.pop(0)
        if name != "layer_with_weights-0":
            logger.warning(f"Skipping unknown weight name: {full_name}")
            continue

        logger.info("Loading TF weight {} with shape {}".format(full_name, shape))

        name = split_name.pop(0)
        if name == "layer_with_weights-0":
            # transformer encoder

            pointer = getattr(pointer, "bert")
            trace.append("bert")

            name = split_name.pop(0)
            if name.startswith("layer_with_weights"):
                layer_num = int(name.split("-")[-1])
                if layer_num <= 2:
                    # embedding layers
                    if layer_num == 0:
                        trace.extend(["embeddings", "word_embeddings"])
                        pointer = getattr(pointer, "embeddings")
                        pointer = getattr(pointer, "word_embeddings")
                    elif layer_num == 1:
                        trace.extend(["embeddings", "position_embeddings"])
                        pointer = getattr(pointer, "embeddings")
                        pointer = getattr(pointer, "position_embeddings")
                    else:
                        trace.extend(["embeddings", "token_type_embeddings"])
                        pointer = getattr(pointer, "embeddings")
                        pointer = getattr(pointer, "token_type_embeddings")
                    trace.append("weight")
                    pointer = getattr(pointer, "weight")
                elif layer_num == 3:
                    # embedding LayerNorm
                    trace.extend(["embeddings", "LayerNorm"])
                    pointer = getattr(pointer, "embeddings")
                    pointer = getattr(pointer, "LayerNorm")
                elif layer_num >= 4 and layer_num < config.num_hidden_layers + 4:
                    # encoder layers
                    trace.extend(["encoder", "layer", str(layer_num - 4)])
                    pointer = getattr(pointer, "encoder")
                    pointer = getattr(pointer, "layer")
                    pointer = pointer[layer_num - 4]
                elif layer_num == config.num_hidden_layers + 4:
                    # pooler layer
                    trace.extend(["pooler", "dense"])
                    pointer = getattr(pointer, "pooler")
                    pointer = getattr(pointer, "dense")
        elif name == "layer_with_weights-1":
            # next sentence prediction

            trace.extend(["cls", "seq_relationship"])
            pointer = getattr(pointer, "cls")
            pointer = getattr(pointer, "seq_relationship")
        elif name == "layer_with_weights-2":
            # masked lm

            trace.extend(["cls", "predictions"])
            pointer = getattr(pointer, "cls")
            pointer = getattr(pointer, "predictions")

            name = split_name.pop(0)
            if name == "dense":
                trace.extend(["transform", "dense"])
                pointer = getattr(pointer, "transform")
                pointer = getattr(pointer, "dense")
            elif name == "layer_norm":
                trace.extend(["transform", "LayerNorm"])
                pointer = getattr(pointer, "transform")
                pointer = getattr(pointer, "LayerNorm")
            elif name == "embedding_table":
                trace.extend(["decoder", "weight"])
                pointer = getattr(pointer, "decoder")
                pointer = getattr(pointer, "weight")
            elif name == "output_bias.Sbias":
                trace.extend(["bias"])
                pointer = getattr(pointer, "bias")
        else:
            logger.warning(f"Skipping unknown weight name: {full_name}")
            continue

        # iterate over the rest depths
        for name in split_name:
            if name == "_attention_layer":
                # self-attention layer
                trace.extend(["attention", "self"])
                pointer = getattr(pointer, "attention")
                pointer = getattr(pointer, "self")
            elif name == "_attention_layer_norm":
                # output attention norm
                trace.extend(["attention", "output", "LayerNorm"])
                pointer = getattr(pointer, "attention")
                pointer = getattr(pointer, "output")
                pointer = getattr(pointer, "LayerNorm")
            elif name == "_attention_output_dense":
                # output attention dense
                trace.extend(["attention", "output", "dense"])
                pointer = getattr(pointer, "attention")
                pointer = getattr(pointer, "output")
                pointer = getattr(pointer, "dense")
            elif name == "_intermediate_dense":
                # attention intermediate dense
                trace.extend(["intermediate", "dense"])
                pointer = getattr(pointer, "intermediate")
                pointer = getattr(pointer, "dense")
            elif name == "_output_dense":
                # output dense
                trace.extend(["output", "dense"])
                pointer = getattr(pointer, "output")
                pointer = getattr(pointer, "dense")
            elif name == "_output_layer_norm":
                # output dense
                trace.extend(["output", "LayerNorm"])
                pointer = getattr(pointer, "output")
                pointer = getattr(pointer, "LayerNorm")
            elif name == "_key_dense":
                # attention key
                trace.append("key")
                pointer = getattr(pointer, "key")
            elif name == "_query_dense":
                # attention query
                trace.append("query")
                pointer = getattr(pointer, "query")
            elif name == "_value_dense":
                # attention value
                trace.append("value")
                pointer = getattr(pointer, "value")
            elif name == "dense":
                # attention value
                trace.append("dense")
                pointer = getattr(pointer, "dense")
            elif name in ["bias", "beta"]:
                # norm biases
                trace.append("bias")
                pointer = getattr(pointer, "bias")
            elif name in ["kernel", "gamma"]:
                # norm weights
                trace.append("weight")
                pointer = getattr(pointer, "weight")
            elif name == ".ATTRIBUTES":
                # full variable name ends with .ATTRIBUTES/VARIABLE_VALUE
                break
            else:
                logger.warning(f"Skipping unknown weight name: {full_name}")

        array = tf.train.load_variable(tf_checkpoint_path, full_name)

        # for certain layers reshape is necessary
        trace = ".".join(trace)
        if re.match(r"(\S+)\.attention\.self\.(key|value|query)\.(bias|weight)", trace) or \
           re.match(r"(\S+)\.attention\.output\.dense\.weight", trace):
            array = array.reshape(pointer.data.shape)
        if "kernel" in full_name:
            array = array.transpose()

        if pointer.shape == array.shape:
            pointer.data = torch.from_numpy(array)
        else:
            raise ValueError(
                f"Shape mismatch in layer {full_name}: "
                f"Model expects shape {pointer.shape} but layer contains shape: {array.shape}"
            )

        logger.info(f"Successfully set variable {full_name} to PyTorch layer {trace}")

    return model


def convert_tf2_checkpoint_to_pytorch(tf_checkpoint_path, config_path, pytorch_dump_path):
    # Initialize PyTorch model
    logger.info(f"Loading model based on config from {config_path}...")
    config = BertConfig.from_json_file(config_path)
    model = BertForPreTraining(config)

    # Load weights from tf checkpoint
    logger.info(f"Loading weights from checkpoint {tf_checkpoint_path}...")
    load_tf2_weights_in_bert(model, tf_checkpoint_path, config)

    # Save pytorch-model
    logger.info(f"Saving PyTorch model to {pytorch_dump_path}...")
    torch.save(model.state_dict(), pytorch_dump_path)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--tf_checkpoint_path", type=str, required=True, help="Path to the TensorFlow 2.x checkpoint path."
    )
    parser.add_argument(
        "--config_file",
        type=str,
        required=True,
        help="The config json file corresponding to the BERT model. This specifies the model architecture.",
    )
    parser.add_argument(
        "--pytorch_dump_path",
        type=str,
        required=True,
        help="Path to the output PyTorch model (must include filename).",
    )
    args = parser.parse_args()
    convert_tf2_checkpoint_to_pytorch(args.tf_checkpoint_path, args.config_file, args.pytorch_dump_path)



# Copyright (c) 2021 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pickle
import shutil
import tempfile
import unittest

from transformers.models.bert.tokenization_bert import WordpieceTokenizer

from sudachitra import BertSudachipyTokenizer, SudachipyWordTokenizer
from sudachitra.tokenization_bert_sudachipy import VOCAB_FILES_NAMES, WORD_FORM_TYPES


class BertSudachipyTokenizationTest(unittest.TestCase):

    tokenizer_class = BertSudachipyTokenizer

    def setUp(self):
        super().setUp()
        self.tmpdirname = tempfile.mkdtemp()

        vocab_tokens = [
            "[UNK]",
            "[CLS]",
            "[SEP]",
            "こんにちは",
            "こん",
            "にちは",
            "ばんは",
            "##こん",
            "##にちは",
            "##ばんは",
            "世界",
            "##世界",
            "、",
            "##、",
            "。",
            "##。",
        ]

        self.vocab_file = os.path.join(self.tmpdirname, VOCAB_FILES_NAMES["vocab_file"])
        with open(self.vocab_file, "w", encoding="utf-8") as vocab_writer:
            vocab_writer.write("".join([x + "\n" for x in vocab_tokens]))

    def tearDown(self):
        shutil.rmtree(self.tmpdirname)

    def test_pickle_sudachipy_tokenizer(self):
        tokenizer = self.tokenizer_class(
            self.vocab_file,
            do_lower_case=False,
            do_word_tokenize=True,
            subword_tokenizer_type='wordpiece'
        )
        self.assertIsNotNone(tokenizer)

        text = "こんにちは、世界。\nこんばんは、世界。"
        tokens = tokenizer.tokenize(text)
        self.assertListEqual(tokens, ["こんにちは", "、", "世界", "。", "こん", "##ばんは", "、", "世界", "。"])
        self.assertListEqual(tokenizer.convert_tokens_to_ids(tokens), [3, 12, 10, 14, 4, 9, 12, 10, 14])

        filename = os.path.join(self.tmpdirname, "tokenizer.bin")
        with open(filename, "wb") as handle:
            pickle.dump(tokenizer, handle)

        with open(filename, "rb") as handle:
            tokenizer_new = pickle.load(handle)

        tokens_loaded = tokenizer_new.tokenize(text)

        self.assertListEqual(tokens, tokens_loaded)

    def test_sudachipy_tokenizer_small(self):
        try:
            tokenizer = SudachipyWordTokenizer(dict_type="small")
        except ModuleNotFoundError:
            return

        self.assertListEqual(
            list(map(WORD_FORM_TYPES['surface'],
                     tokenizer.tokenize("appleはsmall辞書に、apple pieはcore辞書に、apple storeはfull辞書に収録されている。"))),
            ["apple", "は", "small", "辞書", "に", "、", "apple", " ", "pie", "は", "core", "辞書", "に", "、",
             "apple", " ", "store", "は", "full", "辞書", "に", "収録", "さ", "れ", "て", "いる", "。"]
        )

    def test_sudachipy_tokenizer_core(self):
        try:
            tokenizer = SudachipyWordTokenizer(dict_type="core")
        except ModuleNotFoundError:
            return

        self.assertListEqual(
            list(map(WORD_FORM_TYPES['surface'],
                     tokenizer.tokenize("appleはsmall辞書に、apple pieはcore辞書に、apple storeはfull辞書に収録されている。"))),
            ["apple", "は", "small", "辞書", "に", "、", "apple pie", "は", "core", "辞書", "に", "、",
             "apple", " ", "store", "は", "full", "辞書", "に", "収録", "さ", "れ", "て", "いる", "。"]
        )

    def test_sudachipy_tokenizer_full(self):
        try:
            tokenizer = SudachipyWordTokenizer(dict_type="full")
        except ModuleNotFoundError:
            return

        self.assertListEqual(
            list(map(WORD_FORM_TYPES['surface'],
                     tokenizer.tokenize("appleはsmall辞書に、apple pieはcore辞書に、apple storeはfull辞書に収録されている。"))),
            ["apple", "は", "small", "辞書", "に", "、", "apple pie", "は", "core", "辞書", "に", "、",
             "apple store", "は", "full", "辞書", "に", "収録", "さ", "れ", "て", "いる", "。"]
        )

    def test_sudachipy_tokenizer_surface(self):
        tokenizer = self.tokenizer_class(self.vocab_file, do_subword_tokenize=False,
                                         word_form_type='surface')

        self.assertListEqual(
            tokenizer.tokenize("appleの辞書形はAppleで正規形はアップルである。"),
            ["apple", "の", "辞書形", "は", "Apple", "で", "正規", "形", "は", "アップル", "で", "ある", "。"]

        )

    def test_sudachipy_tokenizer_dictionary_form(self):
        tokenizer = self.tokenizer_class(self.vocab_file, do_subword_tokenize=False,
                                         word_form_type='dictionary')

        self.assertListEqual(
            tokenizer.tokenize("appleの辞書形はAppleで正規形はアップルである。"),
            ["Apple", "の", "辞書形", "は", "Apple", "で", "正規", "形", "は", "アップル", "だ", "ある", "。"]
        )

    def test_sudachipy_tokenizer_normalized_form(self):
        tokenizer = self.tokenizer_class(self.vocab_file, do_subword_tokenize=False,
                                         word_form_type='normalized')

        self.assertListEqual(
            tokenizer.tokenize("appleの辞書形はAppleで正規形はアップルである。"),
            ["アップル", "の", "辞書形", "は", "アップル", "で", "正規", "形", "は", "アップル", "だ", "有る", "。"]
        )

    def test_sudachipy_tokenizer_dictionary_form_and_surface(self):
        tokenizer = self.tokenizer_class(self.vocab_file, do_subword_tokenize=False,
                                         word_form_type='dictionary_and_surface')

        self.assertListEqual(
            tokenizer.tokenize("appleの辞書形はAppleで正規形はアップルである。"),
            ["Apple", "の", "辞書形", "は", "Apple", "で", "正規", "形", "は", "アップル", "で", "ある", "。"]
        )

    def test_sudachipy_tokenizer_normalized_form_and_surface(self):
        tokenizer = self.tokenizer_class(self.vocab_file, do_subword_tokenize=False,
                                         word_form_type='normalized_and_surface')

        self.assertListEqual(
            tokenizer.tokenize("appleの辞書形はAppleで正規形はアップルである。"),
            ["アップル", "の", "辞書形", "は", "アップル", "で", "正規", "形", "は", "アップル", "で", "ある", "。"]
        )

    def test_sudachipy_tokenizer_unit_a(self):
        try:
            tokenizer = SudachipyWordTokenizer(split_mode="A")
        except ModuleNotFoundError:
            return

        self.assertListEqual(
            list(map(WORD_FORM_TYPES['surface'], tokenizer.tokenize("徳島阿波おどり空港"))),
            ["徳島", "阿波", "おどり", "空港"]
        )

    def test_sudachipy_tokenizer_unit_b(self):
        try:
            tokenizer = SudachipyWordTokenizer(split_mode="B")
        except ModuleNotFoundError:
            return

        self.assertListEqual(
            list(map(WORD_FORM_TYPES['surface'], tokenizer.tokenize("徳島阿波おどり空港"))),
            ["徳島", "阿波おどり", "空港"]
        )

    def test_sudachipy_tokenizer_unit_c(self):
        try:
            tokenizer = SudachipyWordTokenizer(split_mode="C")
        except ModuleNotFoundError:
            return

        self.assertListEqual(
            list(map(WORD_FORM_TYPES['surface'], tokenizer.tokenize("徳島阿波おどり空港"))),
            ["徳島阿波おどり空港"]
        )

    def test_wordpiece_tokenizer(self):
        vocab_tokens = ["[UNK]", "[CLS]", "[SEP]", "こんにちは", "こん", "にちは" "ばんは", "##こん", "##にちは", "##ばんは"]

        vocab = {}
        for (i, token) in enumerate(vocab_tokens):
            vocab[token] = i
        tokenizer = WordpieceTokenizer(vocab=vocab, unk_token="[UNK]")

        self.assertListEqual(tokenizer.tokenize(""), [])

        self.assertListEqual(tokenizer.tokenize("こんにちは"), ["こんにちは"])

        self.assertListEqual(tokenizer.tokenize("こんばんは"), ["こん", "##ばんは"])

        self.assertListEqual(tokenizer.tokenize("こんばんは こんばんにちは こんにちは"), ["こん", "##ばんは", "[UNK]", "こんにちは"])

    def test_sequence_builders(self):
        pass


class BertSudachipyCharacterTokenizationTest(unittest.TestCase):

    tokenizer_class = BertSudachipyTokenizer

    def setUp(self):
        super().setUp()
        self.tmpdirname = tempfile.mkdtemp()

        vocab_tokens = ["[UNK]", "[CLS]", "[SEP]", "こ", "ん", "に", "ち", "は", "ば", "世", "界", "、", "。"]

        self.vocab_file = os.path.join(self.tmpdirname, VOCAB_FILES_NAMES["vocab_file"])
        with open(self.vocab_file, "w", encoding="utf-8") as vocab_writer:
            vocab_writer.write("".join([x + "\n" for x in vocab_tokens]))

    def test_full_tokenizer(self):
        tokenizer = self.tokenizer_class(self.vocab_file, subword_tokenizer_type="character")

        tokens = tokenizer.tokenize("こんにちは、世界。こんばんは、世界。")
        self.assertListEqual(
            tokens, ["こ", "ん", "に", "ち", "は", "、", "世", "界", "。", "こ", "ん", "ば", "ん", "は", "、", "世", "界", "。"]
        )
        self.assertListEqual(
            tokenizer.convert_tokens_to_ids(tokens), [3, 4, 5, 6, 7, 11, 9, 10, 12, 3, 4, 8, 4, 7, 11, 9, 10, 12]
        )

    def test_sequence_builders(self):
        pass



# Copyright (c) 2021 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import BertSudachipyTokenizer


VOCAB_FILES_NAMES = {"vocab_file": "vocab.txt"}

# TODO: set official URL
PRETRAINED_VOCAB_FILES_MAP = {
    "vocab_file": {
        "megagonlabs/electra-base-ud-japanese": "https://.../vocab.txt",
    }
}

PRETRAINED_POSITIONAL_EMBEDDINGS_SIZES = {
    "megagonlabs/electra-base-ud-japanese": 512,
}

PRETRAINED_INIT_CONFIGURATION = {
    "megagonlabs/electra-base-ud-japanese": {
        "do_lower_case": False,
        "word_tokenizer_type": "sudachipy",
        "subword_tokenizer_type": "pos_substitution",
    },
}


class ElectraSudachipyTokenizer(BertSudachipyTokenizer):

    vocab_files_names = VOCAB_FILES_NAMES
    pretrained_vocab_files_map = PRETRAINED_VOCAB_FILES_MAP
    max_model_input_sizes = PRETRAINED_POSITIONAL_EMBEDDINGS_SIZES
    pretrained_init_configuration = PRETRAINED_INIT_CONFIGURATION



# Copyright (c) 2021 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import tensorflow_datasets as tfds
from bunkai import Bunkai
from typing import List


TARGET_DATASETS = ['train', 'validation', 'test']

START_ARTICLE_DELIMITER = '_START_ARTICLE_'
START_PARAGRAPH_DELIMITER = '_START_PARAGRAPH_'
NEW_LINE_DELIMITER = '_NEWLINE_'


def split_article(article_text: str) -> List[List[str]]:
    """
    Splits an article into paragraphs.

    Args: article_text (str): An article in wikipedia.

    Returns:
        List[List[str]]: List of paragraphs containing sentences
    """
    paragraphs = []
    lines = article_text.split('\n')
    for i in range(2, len(lines), 2):
        if lines[i-1] == START_PARAGRAPH_DELIMITER:
            paragraphs.append(lines[i].split(NEW_LINE_DELIMITER))

    return paragraphs


def main():
    args = get_args()

    ds, ds_info = tfds.load(name='wiki40b/ja', split=['train', 'validation', 'test'], with_info=True)

    bunkai = Bunkai()

    for line in tfds.as_dataframe(ds[TARGET_DATASETS.index(args.target)], ds_info).itertuples():
        paragraphs = split_article(line.text.decode('utf-8'))
        print(START_ARTICLE_DELIMITER)
        for paragraph in paragraphs:
            print(START_PARAGRAPH_DELIMITER)
            for sentences in paragraph:
                for sentence in bunkai(sentences):
                    if sentence:
                        print(sentence)


def get_args():
    parser = argparse.ArgumentParser(description='Download and parse target dataset.')
    parser.add_argument('-t', '--target', choices=['train', 'validation', 'test'], help='Target dataset.')

    args = parser.parse_args()

    return args


if __name__ == '__main__':
    main()
